%{
#include "dynare2.h"
#include "moments.h"

  int yyerror(char *);
  int yylex(void);
  void set_deriv_order(char *);

  extern FILE *f_out;
  extern int yylineno;
  int symblist_flag;
  extern model_struct sm;
  extern Moments moments;
%}

%union {
  char *string;
  int  integer;
  double dvalue;
  struct s_node* result;
      }

%token  END ENDVAL INITVAL MODEL PERIODS SHOCKS VALUES VAR VAREXO VARDEF EQUAL_SIGN
%token VARF VARP DEL SUM PROD TO MSHOCKS PARAMETERS DYN2VEC RPLOT
%token DO TO ENDO PROD BY DOLLAR STEADY VCOV ORDER STDERR
%left PLUS MINUS 
%left TIMES DIVIDE 
%left UMINUS
%right POWER 
%nonassoc FACTORIAL 
%token EXP LOG LOG10 LN SIN COS TAN ASIN ACOS ATAN SINH COSH TANH ASINH ACOSH ATANH SQRT 
%token <string> INUMBER DNUMBER NAME POUND EOL VINDEX
%type <dvalue> value
%type <result> expression var_exp
%%

 statement_list: statement
      | statement_list statement 
      ;
 statement: periods
      | var        
      | varexo      
      | vardef
      | model 
      | parameters
      | set_params
      | initval
      | shocks
      | vcov
      | order
      ;

 periods : PERIODS INUMBER ';'
      | PERIODS '=' INUMBER ';'
      ;

 var : VAR {symblist_flag=1;} symblist ';' {init_var();}
      ;

 varexo : VAREXO {symblist_flag=0;} symblist ';' {init_varexo();}
      ;

 vardef : VARDEF {symblist_flag=2;} symblist ';'
      ;

 symblist : symblist NAME '-' NAME {add_symbol_range($2,$4,symblist_flag);}
      | symblist ',' NAME '-' NAME {add_symbol_range($3,$5,symblist_flag);}
      | symblist NAME              {add_symbol($2,symblist_flag);}
      | symblist ',' NAME          {add_symbol($3,symblist_flag);}
      | NAME '-' NAME              {add_symbol_range($1,$3,symblist_flag);}
      | NAME                       {add_symbol($1,symblist_flag);}
      ; 

parameters : PARAMETERS {symblist_flag=3;} symblist ';' {init_param();}
      ;

 model : MODEL ';' equation_list END ';'
      ;

 equation_list : equation_list equation  {sm.eq_nbr++;}
      | equation                         {sm.eq_nbr++;}
      ;

 equation : expression EQUAL_SIGN expression ';' 
                                  {add_result(sm.eq_nbr,add_node(MINUS,$1,$3));}
      | expression ';'
                                  {add_result(sm.eq_nbr,$1);}
      ;
 expression : '(' expression ')' { $$ = $2;}
      | expression PLUS expression {$$=add_node(PLUS,$1,$3);}
      | expression MINUS expression {$$=add_node(MINUS,$1,$3);}
      | expression TIMES expression {$$=add_node(TIMES,$1,$3);}
      | expression DIVIDE expression {$$=add_node(DIVIDE,$1,$3);}
      | expression POWER expression {$$=add_node(POWER,$1,$3);}
      | MINUS expression %prec UMINUS {$$=add_node(MINUS,0,$2);}
      | SIN '(' expression ')' {$$=add_node(SIN,0,$3);}
      | COS '(' expression ')' {$$=add_node(COS,0,$3);}
      | TAN '(' expression ')' {$$=add_node(TAN,0,$3);}
      | EXP '(' expression ')' {$$=add_node(EXP,0,$3);}
      | LOG '(' expression ')' {$$=add_node(LOG,0,$3);}
      | LOG10 '(' expression ')' {$$=add_node(LOG10,0,$3);}
      | LN  '(' expression ')' {$$=add_node(LOG,0,$3);}
      | ASIN '(' expression ')' {$$=add_node(ASIN,0,$3);}
      | ACOS '(' expression ')' {$$=add_node(ACOS,0,$3);}
      | ATAN '(' expression ')' {$$=add_node(ATAN,0,$3);}
      | SINH '(' expression ')' {$$=add_node(SINH,0,$3);}
      | COSH '(' expression ')' {$$=add_node(COSH,0,$3);}
      | TANH '(' expression ')' {$$=add_node(TANH,0,$3);}
      | ASINH '(' expression ')' {$$=add_node(ASINH,0,$3);}
      | ACOSH '(' expression ')' {$$=add_node(ACOSH,0,$3);}
      | ATANH '(' expression ')' {$$=add_node(ATANH,0,$3);}
      | SQRT '(' expression ')' {$$=add_node(SQRT,0,$3);}
      | var_exp
      | INUMBER {$$=add_constant(INUMBER,$1);}
      | DNUMBER {$$=add_constant(DNUMBER,$1);}
      ;

 var_exp : NAME {$$ = add_variable($1,0,0);}
      | NAME '(' PLUS INUMBER ')' 
             {int lag; sscanf($4,"%d",&lag);$$ = add_variable($1,lag,0);}
      | NAME '(' MINUS INUMBER ')'
             {int lag; sscanf($4,"%d",&lag);$$ = add_variable($1,-lag,0);}
      | NAME '(' INUMBER ')'
             {int lag; sscanf($3,"%d",&lag);$$ = add_variable($1,lag,0);}
      ;

 set_params : NAME EQUAL_SIGN DNUMBER ';' {set_param($1,$3);}
      | NAME EQUAL_SIGN INUMBER ';' {set_param($1,$3);}
      | NAME EQUAL_SIGN MINUS DNUMBER ';' {set_param($1,$4,-1);}
      | NAME EQUAL_SIGN MINUS INUMBER ';' {set_param($1,$4,-1);}
      ;

 initval : INITVAL ';' init_list END ';'
      ;

 shocks: SHOCKS ';' shock_list END ';'
      ;

 init_list : set_expr
      | init_list set_expr
      ;

 set_expr : NAME EQUAL_SIGN DNUMBER ';' {set_var($1,$3);}
      | NAME EQUAL_SIGN INUMBER ';' {set_var($1,$3);}
      | NAME EQUAL_SIGN MINUS DNUMBER ';' {set_var($1,$4,-1);}
      | NAME EQUAL_SIGN MINUS INUMBER ';' {set_var($1,$4,-1);}
      ;

 vcov: VCOV {moments=Moments(sm.exo_nbr);} EQUAL_SIGN '[' mat_list ']' ';'
      ;

 mat_list : mat_list ';' mat_list1 {moments.row_inc();}
      | mat_list1 {moments.row_inc();}
      ;

 mat_list1 : mat_list1 ',' mat_exp
      | mat_list1 mat_exp
      | mat_exp
      ;

 mat_exp : DNUMBER {moments.push($1);}
      | MINUS DNUMBER {moments.push($2,-1);}
      | INUMBER {moments.push($1);}
      | MINUS INUMBER {moments.push($2,-1);}
      | NAME {moments.push(get_param($1));}
      ;  

 order : ORDER EQUAL_SIGN INUMBER ';' { set_deriv_order($3);} 
      ; 

 shock_list : shock_list shock_elem
      | shock_elem
      ;

 shock_elem : VAR NAME ';' STDERR value ';' {moments.set_variance($2,pow($5,2));}
            | VAR NAME EQUAL_SIGN value ';' {moments.set_variance($2,$4);}
            ;

 value: DNUMBER {sscanf($1,"%lf",&$$);}
      | MINUS DNUMBER {moments.push($2,-1);}
      | INUMBER {moments.push($1);}
      | MINUS INUMBER {moments.push($2,-1);}
      | NAME {moments.push(get_param($1));}
      ;  

%%

int yyerror (char *s)
     
{
  fprintf (stderr, "%s at line %d\n", s, yylineno);
  return 1;
}



