#include "dynare2.h"
#include <mtl/lu.h>

extern struct model_struct sm;
extern vec_double ys;

void nlsolve(vec_double & y)
{
  
  vec_double z(sm.endo_nbr), dz(sm.endo_nbr);
  vec_int p(sm.endo_nbr);
  
  double err = 1.0;
  while (1)
    {
      double err = abs(z[mtl::max_abs_index(z)]);
      cerr << "Err: " << err << "\n";
      if (err < 1e-6)
	{
	  break;
	}

      MatA jacob(sm.endo_nbr,sm.endo_nbr);
      jacobian(z, jacob);

      mtl::lu_factor(jacob, p);
      mtl::lu_solve(jacob, p, mtl::scaled(z,-1.0), dz);
      
      mtl::add(dz,ys);
    }
}

int lnsrch(vec_double & xold, double fold,vec_double & g, vec_double & p, double stpmax, vec_double x, double f, vec_double fvec)
{
  double alf=1e-4;

  int n = xold.size();

  double sum = sqrt(sum_squares(p));
  if ( sum > stpmax)
    {
      scale(p, stpmax/sum);
    }
  
  double slope=dot(g,p);


  double test = 0;
  for (vec_double::iterator i = p.begin(), vec_double::iterator j=xold.begin(); i != p.end(); ++i, ++j)
    {
      double tmp = abs(*i) / ((abs(*j) > 1) ? *j : 1); 
      if ( tmp > test )
	{
	  test = tmp;
	}
    }
  double alamin = tolx / test;

  double alam = 1;
  while (1)
    {
      if (alam < alamin)
	{
	  return 1;
	}
      double x = add(xold,scaledp,alam);
      //      vec_double fvec = _ff(x);
      double f =0.5*dot(fvec,fvec);
      if (f <= fold+alf*alam*slope)
	{
	  return 0;
	}
      double tmplam, f2, fold2, alam2;
      if (alam == 1)
	{
	  tmplam = -slope/(2*(f-fold-slope));
	}
      else
	{
	  double rhs1 = f-fold-alam*slope;
	  double rhs2 = f2-fold2-alam2*slope;
	  double a = (rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
	  double b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
	  if (a == 0)
	    {
	      tmplam = -slope/(2*b);
	    }
	  else
	    {
	      double disc = b*b-3*a*slope;
	      if (disc < 0)
		{
		  cerr << "ERROR: Roundoff problem in nlsearch\n";
		  exit(1);
		}
	      else
		{
		  tmplam=(-b+sqrt(disc))/(3*a);
		}
	    }
	  if (tmplam > 0.5*alam)
	    {
	      tmplam = 0.5*alam;
	    }
	}
      alam2=alam;
      f2=f;
      fold2=fold;
      alam=max(tmplam,0.1*alam);
    }
}


